"""
In principle, one could automate the process for merging data annotations
between binding sites that overlap and need to be merged together. For now, this
function and file does nothing, and merging is handled in other ways.
"""


def generate_merge_func(_):
    """
    Returns the identity function
    :param _: the inputt paramter is ignored

    """
    return lambda x: x



"""
This file contains functions for loading data from the rbpdb database.

"""

import math

from .config import (
    ANNOTATION_COLUMN_DELIMITER,
    RBPDB_MOTIF_N_REPEAT_REQ,
    RBPDB_MOTIF_PWM_LETTER_STRENGTH,
    RBPDB_PATH,
)
from .picklify import picklify
from .pwm_scan import (
    get_human_seq,
    motif_to_pwm,
    pwm_degree_of_freedom,
    pwm_scan,
    pwm_scan_naive_brute_force,
    str_to_pwm,
)

rbpdb_all_column_names = [
    "protein_id",
    "annotation_id",
    "creation_date",
    "update_date",
    "gene_name",
    "gene_description",
    "species",
    "taxID",
    "domains",
    "aliases",
    "flag_protein",
    "flag_notes",
    "some_other_id",
    "experimental_id",
    "PUBMED_ID",
    "exp_type",
    "notes",
    "seq_motif",
    "selex_file",
    "aligned_selex_file",
    "aligned_motif_file",
    "PWM_file",
    "PFM_file",
    "logo_file",
    "secondary_structure",
    "in_vivo_notes",
    "in_vivo_file",
    "flag_experimental",
]

rbpdb_all_column_descriptions = rbpdb_all_column_names
rbpdb_columns_of_interest = [
    0,
    1,
    3,
    4,
    5,
    8,
    9,
    13,
    14,
    15,
    16,
    17,
    22,
    24,
    25,
]
rbpdb_default_label_index = [8]

RBPDB_DEFAULT_MOUSE_OVER_INDEX = 9

rbpdb_column_names = [
    rbpdb_all_column_names[i] for i in rbpdb_columns_of_interest
]
rbpdb_column_descriptions = [
    rbpdb_all_column_descriptions[i] for i in rbpdb_columns_of_interest
]


def generate_rbpdb_experiment_to_columns():
    """
    Converts RBPDB file format to a useful one.
    I don't remember what columns means, but this is a one-time preprocessing
    step needed to extract data out of the RBPDB data.

    """
    rbpdb_experiment_file_path = (
        f"{RBPDB_PATH}/RBPDB_v1.3.1_experiments_human_2012-11-21.tdt"
    )
    experiment_id_to_columns_dict = {}
    with open(rbpdb_experiment_file_path) as handle:
        line = handle.readline()
        while line:
            columns = line.split("\t")
            # Here we expect the columns to be:
            # experimental_id, PUBMED_ID, exp_type, notes, seq_motif,
            # selex_file, aligned_selex_file,
            # aligned_motif_file, PWM_file, PFM_file, logo_file,
            # secondary_structure, in_vivo_notes, in_vivo_file, flag
            if columns[14] == "1":
                # The flag means this data is unreliable, according to the RBPDB
                # Readme files
                line = handle.readline()
                continue
            experimental_id = columns[0]
            experiment_id_to_columns_dict[experimental_id] = columns
            line = handle.readline()
    return experiment_id_to_columns_dict


def generate_rbpdb_protein_to_experiment_id():
    """
    Preprocesses RBPDB files to generate a dictionary mapping protein IDs to
    the experiment ID that investigates their binding data.
    Run once in initialization of RNPFind
    """
    rbpdb_protein_experiment_file_path = (
        f"{RBPDB_PATH}/RBPDB_v1.3.1_protExp_human_2012-11-21.tdt"
    )
    protein_id_to_experimental_ids_dict = {}
    with open(rbpdb_protein_experiment_file_path) as handle:
        line = handle.readline()
        while line:
            columns = line.split("\t")
            # Here we expect the columns to be:
            # protein_id, experiment_id, homolog, unique_id
            protein_id = columns[0]
            experimental_id = columns[1]
            protein_id_to_experimental_ids_dict[
                protein_id
            ] = protein_id_to_experimental_ids_dict.get(protein_id, []) + [
                experimental_id
            ]
            line = handle.readline()
    return protein_id_to_experimental_ids_dict


def generate_rbpdb_experimental_to_pwm(letter_strength, n_repeat_req):
    """
    Processes RBPDB files to generate and return a dictionary mapping RBPDB
    experiment IDs to PWM data structures representing the binding motif of the
    RBP under investigation in the experiment of interest.
    :param letter_strength: the number of times more likely a base is considered
        to be when represented by the motif string at a particular position.
    :param n_repeat_req: specifies the assumption to use for the value of n when
        RBPDB files specify a motif such as "(ATG)n", which to us humans means
        a number of repeats of "ATG". This parameter either makes n_repeat_req
        the minimum number of motif repeats for the binding motif, or makes it
        exactly n_repeat_req. I'm not too sure.

    """
    rbpdb_experiment_file_path = (
        f"{RBPDB_PATH}/RBPDB_v1.3.1_experiments_human_2012-11-21.tdt"
    )
    rbpdb_pfm_file_directory = f"{RBPDB_PATH}/rbpdb-human-pfm-matrices/"
    experimental_to_pwm_dict = {}
    with open(rbpdb_experiment_file_path) as handle:
        line = handle.readline()
        while line:
            columns = line.split("\t")
            # Here we expect the columns to be:
            # experimental_id, PUBMED_ID, exp_type, notes, seq_motif,
            # selex_file, aligned_selex_file,
            # aligned_motif_file, PWM_file, PFM_file, logo_file,
            # secondary_structure, in_vivo_notes, in_vivo_file, flag
            if columns[14] == "1":
                # The flag means this data is unreliable, according to the RBPDB
                # Readme files
                line = handle.readline()
                continue

            experimental_id = columns[0]

            assert len(experimental_id) > 0
            pfm_file = columns[9]
            seq_motifs = columns[4]
            pwms = []
            if pfm_file != "\\N":
                pfm_file_path = rbpdb_pfm_file_directory + pfm_file
                with open(pfm_file_path) as pfm_file_handle:
                    raw_pwm_str = pfm_file_handle.read()
                pwm = str_to_pwm(raw_pwm_str, is_transpose=True)
                pwms += [pwm]
            elif seq_motifs not in ("\\N", ""):
                # This experiment still generated some useful data
                seq_motifs = seq_motifs.split(";")
                i = 0
                while i != len(seq_motifs):
                    seq_motif = seq_motifs[i]
                    while ")(" in seq_motif:
                        repeat_end = seq_motif.find(")(")
                        assert seq_motif[repeat_end] == ")"
                        repeat_start = repeat_end
                        while seq_motif[repeat_start] != "(":
                            repeat_start -= 1

                        number_start = repeat_end + 2
                        assert (
                            seq_motif[number_start].isdigit()
                            or seq_motif[number_start] == "n"
                        )
                        number_end = number_start
                        while seq_motif[number_end] != ")":
                            number_end += 1

                        # deal with cases where the number of repeats is
                        # "15-30". Take minimum to be conservative.
                        # Note that most cases would be a single number like
                        # "15".
                        num_of_repeats = min(
                            [
                                int(s)
                                if s != "n"
                                else math.ceil(
                                    n_repeat_req
                                    / (repeat_end - repeat_start - 1)
                                )
                                for s in seq_motif[
                                    number_start:number_end
                                ].split("-")
                            ]
                        )

                        seq_motif = seq_motif.replace(
                            seq_motif[repeat_start : number_end + 1],
                            seq_motif[repeat_start + 1 : repeat_end]
                            * num_of_repeats,
                        )

                    maketrans = str.maketrans
                    all_letters = "wruysn"
                    upper_map = maketrans(all_letters, all_letters.upper())
                    seq_motif = seq_motif.translate(upper_map)
                    if "/" in seq_motif:
                        bracket_start = bracket_end = middle = seq_motif.find(
                            "/"
                        )
                        while seq_motif[bracket_start] != "(":
                            bracket_start -= 1
                        while seq_motif[bracket_end] != ")":
                            bracket_end += 1
                        seq_motif_1 = seq_motif.replace(
                            seq_motif[bracket_start : bracket_end + 1],
                            seq_motif[bracket_start + 1 : middle],
                        )
                        seq_motif_2 = seq_motif.replace(
                            seq_motif[bracket_start : bracket_end + 1],
                            seq_motif[middle + 1 : bracket_end],
                        )
                        seq_motifs += [seq_motif_1, seq_motif_2]
                    else:
                        pwm = motif_to_pwm(
                            seq_motif, letter_strength=letter_strength
                        )
                        pwms += [pwm]
                    i += 1

            # Now we have the raw text, we convert it to pwm and add to
            # dictionary
            experimental_to_pwm_dict[experimental_id] = pwms
            line = handle.readline()

    return experimental_to_pwm_dict


def rbpdb_data_load(rna_info):
    """
    Returns a Generator(/Iterator?) that represent binding sites loaded from the
    RBPDB database on an RNA molecule of interest.
    :param rna_info: a dictionaru containing the chromosome number, start, and
        end coordinate of the RNA molecule of interest (in hg38).
    :param out: if specified, redirects progress status to the specified
        function instead of stdout. (Default value = None)

    """
    rbpdb_protein_file_path = (
        f"{RBPDB_PATH}/RBPDB_v1.3.1_proteins_human_2012-11-21.tdt"
    )
    letter_strength = RBPDB_MOTIF_PWM_LETTER_STRENGTH
    n_repeat_req = RBPDB_MOTIF_N_REPEAT_REQ
    rna_seq = get_human_seq(rna_info)

    experiment_id_to_pwm_dict = picklify(
        generate_rbpdb_experimental_to_pwm, letter_strength, n_repeat_req
    )
    protein_id_to_experimental_ids_dict = picklify(
        generate_rbpdb_protein_to_experiment_id
    )
    experiment_id_to_columns_dict = picklify(
        generate_rbpdb_experiment_to_columns
    )
    with open(rbpdb_protein_file_path) as handle:
        _ = handle.readline().strip().split("\t")
        # columns here is expected to have the following information in the
        # following order:
        # protein_id, annotation_id, creation_date, update_date, gene_name,
        # gene_description, species, taxID, domains, aliases, flag, flag_notes,
        # some_other_id
        protein_columns = handle.readline().replace("\n", "").split("\t")
        while protein_columns != [""]:
            assert len(protein_columns) == 13
            # We only care about human RBPs for now.
            if protein_columns[10] == "0":
                protein_columns = (
                    handle.readline().replace("\n", "").split("\t")
                )
                continue
            rbp = protein_columns[4]
            protein_id = protein_columns[0]

            if protein_id not in protein_id_to_experimental_ids_dict:
                # No experiments associated. So no data to be had
                protein_columns = (
                    handle.readline().replace("\n", "").split("\t")
                )
                continue

            for experiment_id in protein_id_to_experimental_ids_dict[
                protein_id
            ]:
                assert (
                    experiment_id in experiment_id_to_pwm_dict
                    or experiment_id == "410"
                )
                if experiment_id == "410":
                    continue
                pwms = experiment_id_to_pwm_dict[experiment_id]
                for pwm in pwms:
                    assert len(pwm["A"]) > 0
                    experimental_columns = experiment_id_to_columns_dict[
                        experiment_id
                    ]
                    assert len(experimental_columns) == 15
                    total_columns = protein_columns + experimental_columns
                    annotation = ANNOTATION_COLUMN_DELIMITER.join(
                        [total_columns[i] for i in rbpdb_columns_of_interest]
                    )

                    if pwm_degree_of_freedom(pwm) >= 2048:
                        # experimentally shown that by this point naive brute
                        # force is faster. Bound could be
                        # reduced.
                        sites = pwm_scan_naive_brute_force(rna_seq, pwm)
                    else:
                        sites = pwm_scan(rna_seq, pwm)

                    if not sites:
                        continue

                    for start, end in sites:
                        yield rbp, start, end, annotation

            protein_columns = handle.readline().replace("\n", "").split("\t")



"""
Picklify is a function that works similar to memoization; it is meant for
functions that return a dictionary. Often, such functions will parse a file to
generate a dictionary that maps certain keys to values. To save on such
overhead costs, we "picklify" them the first time they are called (save the
dictionary in a pickle file), and then simply load the dictionary from the
saved pickle files the next time around.

"""
import pickle
from pathlib import Path

from .config import PICKLE_PATH


def picklify(dict_generator, *args, **kwargs):
    """
    Given a function that returns an object such as a dictionary (only dict
    fully supported), returns the dictionary generated by the function. The
    function is only called if it has not been "picklified" (passed as an
    argument to this function) before. Otherwise, its cached dictionary is
    returned instead. Thus getting the dicttionary by wrapping this function
    speeds up the dictionary creation overall.

    Note that this function should not be called by two different functions
    with the same name.

    :param dict_generator: the function which generates a dictionary.
    :param *args: Any args to pass to the dictionary
    :param **kwargs: Any keyword args to pass to the dictionary.
    :returns: dictionary returned by dict_generator().

    """
    # Danger! Never call picklify with functions that have the same name!
    pickle_path = f"{PICKLE_PATH}/{dict_generator.__name__}.pickle"

    try:
        with open(pickle_path, "rb") as pickle_handle:
            dict_to_return = pickle.load(pickle_handle)
    except FileNotFoundError:
        dict_to_return = dict_generator(*args, **kwargs)
        Path(pickle_path).parent.mkdir(parents=True, exist_ok=True)
        with open(pickle_path, "wb") as pickle_handle:
            pickle.dump(
                dict_to_return, pickle_handle, protocol=pickle.HIGHEST_PROTOCOL
            )
    return dict_to_return



"""
Defines functions useful for creating and saving RBP binding sites on a given
RNA transcript, as BED files.

This module is useful for ucsc_visualize data analysis method.
"""
import os
import sys
from pathlib import Path

from .colors import green, red
from .config import GENOME_VERSION
from .data_load_functions import data_load_source_colors
from .load_data import data_source_annotation_to_columns


def populate_binding_sites(
    big_storage, rna_info, data_load_sources, overarching_path, main_rbp="",
):
    """
    Saves binding sites as BED files
    :param big_storage: a Storage instance containing binding sites data for
        RBPs.
    :param rna_info: a dictionary containing the official_name, chr_n, and
        start_coord for the RNA molecule of interest (on the genome)
    :param data_load_sources: An iterable of data load sources from which the
        binding sites are populated in the big_storage
    :param main_rbp: one RBP of interest, to allow for deducing competitive
        and cooperative relationships against
    """

    rna_chr_no = rna_info["chr_n"]
    rna_start_chr_coord = rna_info["start_coord"]

    # TODO: check if this -1 patch is necessary for all data sources or just
    # RBPDB
    displacement = rna_start_chr_coord - 1

    comp_color = red
    coop_color = green
    competitive_threshold_bp = 15
    cooperative_threshold_bp = 56

    for data_load_source in data_load_sources:
        storage = big_storage[data_load_source]

        folder_path = str(Path(overarching_path) / data_load_source) + "/"

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
            os.chmod(folder_path, 0o777)
            print(f"Directory {folder_path} created...", file=sys.stderr)
        else:
            print(
                f"Directory {folder_path} already exists...", file=sys.stderr
            )

        # with open(
        #     Path(overarching_path) / "threshold_config.txt", "w"
        # ) as threshold_config_file:
        #     threshold_config_file.write(
        #         "competitive threshold used: "
        #         + str(competitive_threshold_bp)
        #         + "\n"
        #     )
        #     threshold_config_file.write(
        #         "cooperative threshold used: "
        #         + str(cooperative_threshold_bp)
        #         + "\n"
        #     )

        default_color = data_load_source_colors[data_load_source]

        def coloring_func(
            binding_site, storage=storage, default_color=default_color
        ):
            """
            Defines a coloring function for the print_bed function of the
            Storage instance. This one in particular colors based on
            competitive and cooperative RBPs in comparison with main_rbp. Sites
            that are very close to main_rbp binding sites are colored with
            comp_color (red), and those that are far but not too far are
            colored with coop_color (green). Those that are too far are neutral
            / independent / default.

            :param binding_site: binding site under consideration

            """
            competitive = main_rbp in storage.binds_near(
                binding_site, bp_threshold=competitive_threshold_bp
            )
            cooperative = main_rbp in storage.binds_near(
                binding_site, bp_threshold=cooperative_threshold_bp
            )
            return (
                comp_color
                if competitive
                else coop_color
                if cooperative
                else default_color
            )

        for rbp in storage:
            total_sites = storage[[rbp]].print_bed(
                chr_n=rna_chr_no,
                displacement=displacement,
                end_inclusion=True,
                add_annotation=True,
                include_color=True,
                include_header=False,
                conditional_color_func=coloring_func,
                is_additional_columns=True,
                annotation_to_additional_columns=data_source_annotation_to_columns[
                    data_load_source
                ],
            )

            # filepath = (
            #     rbp
            #     + "_"
            #     + data_load_source
            #     + "_"
            #     + GENOME_VERSION
            #     + "_sites.bed"
            # )
            filepath = f"{rbp}-{data_load_source}-{GENOME_VERSION}-sites.bed"

            filepath = folder_path + filepath

            Path(folder_path).mkdir(parents=True, exist_ok=True)
            with open(filepath, "w") as bed_file:
                bed_file.write(total_sites)

    return overarching_path



"""
Defines the class BindingSites, which represents one RNA molecule and regions
(intervals) where proteins bind on its structure.

BindingSites allows adding binding sites, removing binding sites, querying
binding sites given interval ranges of interest, measuring "correlations"
between pairs of sets of binding sites, printing BED files, and other related
functions.

Note that this project concerns itself with RNAs and proteins, but this class
could very well be used in other similar scenarios (like proteins interacting
with DNA).
"""

from operator import itemgetter

from sortedcontainers import (  # Allow sorted brackets of binding sites
    SortedSet,
)

firstItem = itemgetter(0)
secondItem = itemgetter(1)
firstTwoItems = itemgetter(0, 1)
thirdItem = itemgetter(2)

OVERLAP_CONFLICT = "union"


# Todo: implement a feature that allows belonging relationship to mean overlap
# over the structure...
# Todo: implement a map function
# Todo: Go through the class functions and change replace "-1" default arguments
# with "None" instead


class BindingSites:
    """
    BindingSites allows adding binding sites, removing binding sites, querying
    binding sites given interval ranges of interest, measuring "correlations"
    between pairs of sets of binding sites, printing BED files, and other
    related functions.

    BindingSites can be used in two modes:
        1. when overlap_mode is on, sites are allowed to overlap. This allows
            for queries into "depths" of binding (i.e. how many proteins are
            binding together in a given interval on the RNA).
        2. when overlap_mode is off, sites that overlap are always merged
            together, with no "depth" information maintained. The identity and
            description of binding sites to be merged are also merged and
            associated with the merged binding site

    Note that this project concerns itself with RNAs and proteins, but this
    class could very well be used in other similar scenarios (like proteins
    interacting with DNA).

    The underlying implementation uses an ordered set to keep track of the
    ranges, while adding a range to the set will check for overlaps and deal
    with them.

    This may have been unnecessary, as there are O(nlogn) algorithms that
    produce merged intervals given a list of ranges that can be found online.
    For example:
    https://codereview.stackexchange.com/
                                questions/69242/merging-overlapping-intervals

    However I have implemented this now so I will keep it. However, this may
    allow for simple dynamic additions and deletions with O(logn) each time.

    Major modifications note July 21st 2019:
    Most of the functions defined here have an implicit prerequisite that the
    binding site intervals are non overlapping. As a result,
    the BindingSites.add() function makes sure that the overlapping intervals
    are joined together to form a (possibly) larger interval that includes both
    of the intervals.

    However, now we see that some of the input data might be from experimental
    sources where overlaps are very important (i.e. they represent confidence of
    binding regions because more sequences were identified from that particular
    region). As a result, I am adding a second class variable that keeps track
    of the raw binding sites, so that more functionality can be supported, such
    as finding out the "depth of support" for RBPs binding to a specific
    nucleotide, as well as filters for collapsing the overlapping region based
    on a criteria (e.g. only those with support depth 5 or above, etc.)

    Therefore, as of now, there are two main supported ways of using
    BindingSites:
        1. You add intervals and let BindingSites dynamically take care of
            overlapping intervals for you. Note that as of now, this is a
            default behaviour that always occurs.
        2. You initialize BindingSites with a overlap_mode = True.
            You then add intervals that are highly overlapping and, once
            complete, call the overlap_collapse() function to collapse all the
            intervals as per your specifications. This sets the overlap_mode to
            off. Following this, the representation of the BindingSites to the
            client magically changes to the non-overlapping counterparts and
            enables the functions previously disabled as overlap_mode was
            switched on.

    If overlap_collapse() is called too soon, everything has to be loaded again
    fresh.
    """

    def __init__(self, list_of_sites=None, overlap_mode=False):
        if list_of_sites is None:
            list_of_sites = []
        self.overlap_mode = overlap_mode
        # Just a sorted set underneath
        self.sorted_sites = SortedSet()
        for site in list_of_sites:
            self.add(site)

    def __repr__(self, display_meta=False):
        """Representation of BindingSites objects.

        Show all three elements (start, end, metadata) optionally by setting
        dispMeta = True or alternatively just show the first two tuple elements
        for succinctness
        """
        overlap_add = "OverlapOn" if self.overlap_mode else ""
        if display_meta:
            return self.sorted_sites.__repr__().replace(
                "SortedSet", "BindingSites" + overlap_add
            )

        return (
            SortedSet(map(firstTwoItems, self.sorted_sites))
            .__repr__()
            .replace("SortedSet", "BindingSites" + overlap_add)
        )

    def __str__(self):
        return self.__repr__(display_meta=True)

    def __len__(self):
        return self.sorted_sites.__len__()

    def __iter__(self):
        return self.sorted_sites.__iter__()

    def __getitem__(self, item):
        # Allow for slice selections of elements in BindingSites
        if isinstance(item, slice):
            return BindingSites(self.sorted_sites[item])
        return self.sorted_sites[item]

    @staticmethod
    def is_overlap_ranges(interval_1, interval_2):
        """True iff the ranges (intervals) p and q overlap

        :param p: an interval in the form (start, end) or (start, end, metadata)
        :param q: an interval in the form (start, end) or (start, end, metadata)

        """

        start_1, end_1, *_ = interval_1
        start_2, end_2, *_ = interval_2
        return start_1 <= end_2 and start_2 <= end_1

    @staticmethod
    def _merge_meta(annotation_list, user_merge_func=None):
        """
        Internal function for merging the annotations of multiple
        binding site ranges. The annotations are merged into a tuple of
        annotatins. Some of the annotations passed in may be tuples or lists of
        annotations, too.

        :param annotation_list: A list of annotations to be merged
        :param user_merge_func: if specified, this function is used to merge the
                                metadata instead (Default value = None)

        """

        # TODO: fix the poor style of this function

        # Get all the annotations first from the input list
        new_l = []
        for element in annotation_list:
            if element is None:
                continue
            if isinstance(element, tuple):
                for term in element:
                    new_l.append(term)
            else:
                new_l.append(element)

        if len(new_l) == 0:
            return None

        if len(new_l) == 1:
            return new_l[0]

        # Use user-defined function to merge the list of
        # annotations
        if user_merge_func is not None:
            return user_merge_func(new_l)

        # Otherwise, make a tuple of it, unless its just one element
        new_l_set = set(new_l)

        assert len(new_l_set) > 1

        if len(new_l_set) > 1:
            return tuple(new_l_set)

        return new_l_set.pop()

    @staticmethod
    def _collapse(interval_list, user_merge_func=None):
        """Takes a list of overlapping ranges and collapses them into one

        :param interval_list: A list of intervals to be merged
        :param user_merge_func: if specified, this function is used to merge the
                                metadata (otherwise, metadata is simply joined
                                into a tuple) (Default value = None)

        """

        # assert(not self.overlap_mode)

        if OVERLAP_CONFLICT == "union":
            to_return = (
                min(interval_list, key=firstItem)[0],
                max(interval_list, key=secondItem)[1],
                BindingSites._merge_meta(
                    list(map(thirdItem, interval_list)), user_merge_func
                ),
            )

        elif OVERLAP_CONFLICT == "intersect":
            to_return = (
                max(interval_list, key=firstItem)[0],
                min(interval_list, key=secondItem)[1],
                BindingSites._merge_meta(
                    list(map(thirdItem, interval_list)), user_merge_func
                ),
            )

        return to_return

    def add(self, new_site, user_merge_func=None):
        """Dynamic addition of a range to a sorted set of non-overlapping ranges
        while maintaining the sorted property and merging any produced overlaps.

        May not be the most efficent way of doing this, as _collapse function
        does not take advantage of the sortedness of the ranges.

        :param new_site: the new site to be added, in the form
                         (start, end, metadata) or (start, end).

        :param user_merge_func: If specified, this function is used to merge
                                the metadata of the binding sites in case of
                                overlaps (and a need to merge the sites, due to
                                overlap_mode being off).
                                This function should expect a list of
                                annotations and return one.
                                Otherwise, the default behaviour collates the
                                annotations into a tuple
                                (Default value = None)

        """

        if len(new_site) == 2:
            start, end = new_site
            new_site = (start, end, None)
        elif len(new_site) != 3:
            raise ValueError(
                "Please keep three values in the tuple: "
                + "(start, end, annotation)"
            )

        start, end, _ = new_site
        if not isinstance(start, int) or not isinstance(end, int):
            raise ValueError("Please make sure start and end are integers")

        if start > end:
            raise ValueError(
                "Please make sure the interval end point is greater than the"
                " start point!"
            )

        if self.overlap_mode:
            self.sorted_sites.add(new_site)
            return

        # binary search to find where the new range lies
        start_pos = self.sorted_sites.bisect_left((start, 0))
        end_pos = self.sorted_sites.bisect_left((end, 0))

        # initiate list of ranges that might be merged
        to_merge = [new_site]

        # indices of the sorted set to look at that have the potential for
        # overlapping
        lower = max(0, start_pos - 1)
        higher = min(end_pos + 1, len(self.sorted_sites))

        # This part could be O(n) theoretically but experimentally,
        # (higher-lower) is always strictly less than 5 for this data
        for site in self.sorted_sites[lower:higher]:
            if BindingSites.is_overlap_ranges(site, new_site):
                self.sorted_sites.remove(site)
                to_merge.append(site)

        self.sorted_sites.add(
            BindingSites._collapse(to_merge, user_merge_func)
        )

    def remove(self, site):
        """
        Removes a binding site from an instance of BindingClass.
        :param site: a site to be removed from the BindingClass object. This
                     should exactly match the site that is in the object
                     (if metadata was specified, it should be specified too).
                     This is easier done by obtaining the site using one of the
                     query methods.

        """
        self.sorted_sites.remove(site)

    def dist(self, sites, bp_threshold=30):
        """Checks for correlation between binding sites of two BindingSites.
        Returns a value from 0 to 1.

        WARNING: a.dist(b) and b.dist(a) can give VERY different answers!
        This is because this function checks the "distances" of each of the
        binding sites of one set of BindingSites with all of the other set
        to count the minimal distance and give a scoring based on that.
        Since one of the set of BindingSites could be ubiquitous, the scores
        may vary greatly.

        :param sites: a BindingSites object to compare with.
        :param bp_threshold: the threshold distance value beyond which a site
                             is considered too far from another. The score of
                             proximity is given based on this cutoff value.
                             (Default value = 30)

        """

        if self.overlap_mode:
            raise ValueError(
                "dist() is not supported for BindingSites with"
                " overlap_mode set to True"
            )

        # Todo: break this function into two: one that accepts one site, and
        # another that accepts a BindingSite
        if isinstance(sites, tuple):  # only one tuple input
            start = sites[0]
            end = sites[1]
            pos = self.sorted_sites.bisect_left((start, 0))

            # tuple at the beginning
            if pos == 0:
                dist_end = max(0, self.sorted_sites[pos][0] - end)
                return max(0, 1 - dist_end / bp_threshold)

            # tuple at the end
            if pos == len(self.sorted_sites):
                dist_start = max(0, start - self.sorted_sites[pos - 1][1])
                return max(0, 1 - dist_start / bp_threshold)

            # tuple in the middle
            dist_start = max(0, start - self.sorted_sites[pos - 1][1])
            dist_end = max(0, self.sorted_sites[pos][0] - end)
            # return the closer distance
            return max(0, 1 - min(dist_start, dist_end) / bp_threshold)

        if isinstance(sites, BindingSites):  # a set of tuples given
            cum = 0
            for site in sites:
                cum += self.dist(site, bp_threshold)

            return cum / len(sites)

        # non-supported input type
        print("sites is of type", type(sites))
        raise ValueError(
            "Unsupported type for sites, should be a tuple"
            + " or BindingSites"
        )

    def is_overlap(self, query_site):
        """This checks if an input tuple range overlaps one of those present in
        the set of binding sites stored in self and returns a bool to indicate
        the result.

        :param query_site: a site in the form (start, end, metadata) or just
                            (start, end)

        """

        if self.overlap_mode:
            raise ValueError(
                "isOverlap() is not supported for BindingSites"
                " with overlap_mode set to True"
            )

        start, end, *_ = query_site

        # binary search to find where the query range might lie
        start_pos = self.sorted_sites.bisect_left((start, 0))
        end_pos = self.sorted_sites.bisect_left((end, 0))

        # indices of the sorted set to look at that have the potential for
        # overlapping
        lower = max(0, start_pos - 1)
        higher = min(end_pos + 1, len(self.sorted_sites))

        for site in self.sorted_sites[lower:higher]:
            if BindingSites.is_overlap_ranges(site, query_site):
                return True
        return False

    def nearest_site(self, query_site):
        """This returns the closest range to the input tuple range
        present in the set of binding sites stored in self

        :param query_site: a site in the form (start, end, metadata) or just
                            (start, end)

        """

        if self.overlap_mode:
            raise ValueError(
                "nearestSite() is not supported for BindingSites"
                " with overlap_mode set to True"
            )

        start, end, *_ = query_site
        pos = self.sorted_sites.bisect_left((start, 0))

        # tuple at the beginning
        if pos == 0:
            dist_end = self.sorted_sites[pos][0] - end
            return self.sorted_sites[pos], max(0, dist_end)

        # tuple at the end
        if pos == len(self.sorted_sites):
            dist_start = start - self.sorted_sites[pos - 1][1]
            return self.sorted_sites[pos - 1], max(0, dist_start)

        # tuple in the middle
        dist_start = start - self.sorted_sites[pos - 1][1]
        dist_end = self.sorted_sites[pos][0] - end

        if dist_start > dist_end:
            return self.sorted_sites[pos], max(0, dist_end)

        return self.sorted_sites[pos - 1], max(0, dist_start)

    @staticmethod
    def distance(site_1, site_2):
        """
        Returns the distance between two interval ranges. If they overlap, the
        distance is zero.

        :param site_1: a site in the form (start, end, metadata) or just
                            (start, end)
        :param site_2: a site in the form (start, end, metadata) or just
                            (start, end)

        """
        if BindingSites.is_overlap_ranges(site_1, site_2):
            return 0

        start_1, end_1, *_ = site_1
        start_2, end_2, *_ = site_2
        return min(abs(start_1 - end_2), abs(start_2 - end_1))

    def print(self):
        """
        prints the BindingSite instance
        """
        print(self)

    def len(self):
        """
        returns the number of sites stored in the BindingSite instance
        """
        return len(self)

    def filter_overlap(self, query_range, bp_threshold=0):
        """
        Given a query range, returns a new BindingSite instance with only sites
        that overlap the given range.

        :param query_range: a range/site in the form (start, end) or
                            (start, end, metadata)
        :param bp_threshold:  (Default value = 0)

        """
        if self.overlap_mode:
            raise ValueError(
                "filterOverlap() is not supported for BindingSites"
                " with overlap_mode set to True"
            )

        # Returns all the sites which overlap with the input range query_range
        # given
        start, end, *_ = query_range
        start = start - bp_threshold
        end = end + bp_threshold
        query_range = (start, end)
        start_pos = self.sorted_sites.bisect_left((start, 0))
        end_pos = self.sorted_sites.bisect_left((end, 0))

        # indices of the sorted set to look at that have the potential for
        # overlapping
        lower = max(0, start_pos - 1)
        higher = min(end_pos + 1, len(self.sorted_sites))

        output_binding_sites = BindingSites()
        for site in self.sorted_sites[lower:higher]:
            if BindingSites.is_overlap_ranges(site, query_range):
                output_binding_sites.add(site)
        return output_binding_sites

    def print_bed(
        self,
        name="Generic Binding Site",
        chr_n=1,
        displacement=0,
        end_inclusion=False,
        add_annotation=False,
        include_score=False,
        score_max=1000,
        score_base=1000,
        include_color=False,
        conditional_color_func=None,
        is_bar=False,
        is_additional_columns=False,
        annotation_to_additional_columns=None,
    ):
        """

        :param name:  (Default value = "Generic Binding Site")
        :param chr_n:  (Default value = 1)
        :param displacement:  (Default value = 0)
        :param end_inclusion:  (Default value = False)
        :param add_annotation:  (Default value = False)
        :param include_score:  (Default value = False)
        :param score_max:  (Default value = 1000)
        :param score_base:  (Default value = 1000)
        :param include_color:  (Default value = False)
        :param conditional_color_func:  (Default value = None)
        :param is_bar:  (Default value = False)
        :param is_additional_columns:  (Default value = False)
        :param annotation_to_additional_columns:  (Default value = None)

        """

        # Todo: investigate the reason for add_annotation being an unused
        # argument
        del add_annotation

        # Todo: Possibly remove the functionality for is_bar, it seems
        # misplaced!
        output_str = ""
        if not isinstance(chr_n, str):
            chr_n = "chr" + str(chr_n)
        else:
            chr_n = ("chr" + chr_n) if chr_n[:3] != "chr" else chr_n

        for _tuple in self.sorted_sites:
            start, end, annotation = _tuple
            start, end = (
                displacement + start,
                (displacement + end + (1 if end_inclusion else 0)),
            )

            name_display = name
            to_join = [chr_n, start, end, name_display]

            if include_color and not include_score:
                score = 1000
                to_join.append(score)
            elif include_score:

                assert len(annotation) == 1

                score = float(
                    "".join(
                        filter(
                            lambda k: k.isdigit() or k == "." or k == "-",
                            annotation[0],
                        )
                    )
                )
                score = int(score / score_base * score_max)
                to_join.append(score)

            if include_color and is_bar:
                raise ValueError("Cant be both color and bar!")

            if include_color:
                strand = "+"  # default

                if conditional_color_func is None:
                    color = "0,0,0"  # black
                else:
                    red, green, blue = conditional_color_func(_tuple)
                    color = ",".join(map(str, [red, green, blue]))

                to_join += [strand, start, end, color]

            if is_bar and not include_score:
                raise ValueError("What height for bar?")
            if is_bar:
                strand = "+"  # default
                number_of_bars = 1
                to_join += [strand, name, number_of_bars, score]

            if is_additional_columns:
                to_join += [
                    s.replace(" ", "_") if s else ".'"
                    for s in annotation_to_additional_columns(annotation)
                ]
            output_str += "\t".join(map(str, to_join)) + "\n"

        return output_str

    def return_depth(self, length=-1):
        """
        Returns the density array of binding sites stored.

        Note that binding sites can bind on the 0th nucleotide, and cannot bind
        on the length'th nucleotide

        :param length: a parameter guaranteed to be bigger than the end point
        of any site stored in the BindingSites instance. Specify -1 if unknown.
        (Default value = -1)

        """

        if length == -1:
            if len(self) == 0:
                raise ValueError(
                    "If the BindingSites object is empty, please"
                    " do not call return_depth without "
                    "specifying the length parameter."
                )
            length = max(map(secondItem, self)) + 1

        # Stores 'depth' of support for each nucleotide in the
        # molecule in terms of its chances of being a binding site.
        binding_depth = [0] * length

        for site in self:

            start = site[0]
            end = site[1]

            for nucleotide in range(start, end + 1):  # inclusive
                binding_depth[nucleotide] += 1

        return binding_depth

    def overlap_collapse(
        self, mode, number, in_place=False, annotation_merger=None
    ):
        """Collapses the overlapping ranges to non-overlapping ones, based on
        preset conditions.

        This function will always look at the 'depths' of how much coverage
        support each nucleotide position has and chooses a cutoff point - e.g.
        all nucleotides above depth level of 5 is kept as binding sites and the
        rest are discarded. The cutoff point can  be chosen through multiple
        means.

        The modes supported right now are 'baseCoverNumber', 'TopDepthRatio',
        'TopDepthNumber', 'MinimumDepthNumber', 'TopSitesNumber','TopSitesRatio'

            'baseCoverNumber': Choose cut off based on number of bases that
                should be covered by the selected sites. The most stringent
                cutoff that achieves this criteria is selected, unless not
                possible*.

            'TopDepthRatio': Choose cutoff based on the fraction of highest
                depth coverage that should be supported as binding sites. For
                example, if the deepest coverage provided is 10 and number=0.4,
                then depth coverage of 10,9,8,7,and 6 is counted as binding
                sites and the rest are disregarded.

            'TopDepthNumber': The number of layers of depth from the highest
                depth support that should be selected is input, and the cutoff
                is accordingly selected.

            'MinimumDepthNumber': The cut-off is selected based on the minimum
                depth support each binding site should have.

            'TopSitesNumber': The cut-off is selected such that the top selected
                number of binding sites remains supported. For example, the top
                100 sites may be preserved (from a set of, say, 1000 overlapping
                sites)

            'TopSitesRatio': The cut-off is selected much like above, but the
                ratio of top sites that should be selected is specified instead.
                For example, in the above example, 0.1 could be specified
                instead.

        As of now, calling overlap_collapse() loses all annotation data
        associated with the original range interval data.

        If inPlace is set to True, the BindingSites variable changes and
        collpases, otherwise a new BindingSites variable is generated and
        returned.

        *In general, no nucleotide with support<1 is kept.

        :param mode: a string representing the mode to be used
        :param number: an appropriate numeric argument based on the mode
                       specified
        :param in_place: if True, modifies the instance on which this function
                         is called (and sets overlap_mode to False).
                         Otherwise, creates a new BindingSites instance and
                         returns it (Default value = False)
        :param annotation_merger: If specified, this argument (a function) is
                                  used to merge the annotations of the
                                  overlapping binding sites. Otherwise, the
                                  annotations are merged into a tuple
                                  (Default value = None)

        """
        if not self.overlap_mode:
            print(
                "WARNING: overlap_collapse() called although overlap_mode is"
                " set to off!"
            )

        depth_array = self.return_depth()

        max_depth = max(depth_array)

        if mode == "baseCoverNumber":
            depth_cutoff = -1
            while (
                len(list(filter(lambda k: k > depth_cutoff, depth_array)))
                > number
            ):
                depth_cutoff += 1

            if depth_cutoff == -1:
                # print("WARNING: your baseCoverNumber is impossible to"
                #       " achieve!")
                pass

        elif mode == "TopDepthRatio":
            if not 0 <= number <= 1:
                raise ValueError("Ratio should be between 0 and 1")

            depth_cutoff = max_depth * (1 - number)

        elif mode == "TopDepthNumber":
            depth_cutoff = max_depth - number

        elif mode == "MinimumDepthNumber":
            depth_cutoff = number - 1

        elif mode == "TopSitesNumber":
            raise ValueError("Unimplemented function")
        elif mode == "TopSitesRatio":
            raise ValueError("Unimplemented function")
        else:
            raise ValueError(
                "The mode selected, '" + mode + "' is not" " supported!"
            )

        sites = self.sorted_sites
        depth_cutoff = max(0, depth_cutoff)
        if in_place:
            self.overlap_mode = False
            self.sorted_sites = SortedSet()
            binding_site_to_add_to = self
        else:
            binding_site_to_add_to = BindingSites()

        in_range = False
        start_range = 0
        end_range = 0
        nucleotide = 0
        for nucleotide, depth in enumerate(depth_array):
            if depth > depth_cutoff:
                if not in_range:
                    start_range = nucleotide
                    in_range = True
            else:
                if in_range:
                    end_range = nucleotide - 1
                    in_range = False
                    binding_site_to_add_to.add((start_range, end_range))

        if in_range:
            end_range = nucleotide
            in_range = False
            binding_site_to_add_to.add((start_range, end_range))

        # Add annotations
        for site in sites:
            start, end, annotation = site
            # print(start, end, annotation)
            site, distance = binding_site_to_add_to.nearest_site(site)
            if distance == 0:
                binding_site_to_add_to.add(
                    (start, end, annotation), annotation_merger
                )
        if not in_place:
            return binding_site_to_add_to

        return BindingSites()

    def base_cover(self):
        """
        Returns the number of bases of the RNA molecule that have at least one
        protein bound to it

        (or simply, the length of space covered by the intervals stored in
        BindingSites)
        """
        depth_array = self.return_depth()
        return len(list(filter(lambda k: k > 0, depth_array)))

    def print_wig(
        self,
        chr_no=1,
        displacement=0,
        include_name=False,
        include_description=False,
        name="",
        description="",
        include_header=True,
        length=-1,
    ):
        """Prints a wig file depicting density of binding sites by the RBP.

        Optional parameter length allows for plotting 0 beyond the rightmost
        binding site if needed.

        :param chr_no:  (Default value = 1)
        :param displacement:  (Default value = 0)
        :param include_name:  (Default value = False)
        :param include_description:  (Default value = False)
        :param name:  (Default value = "")
        :param description:  (Default value = "")
        :param include_header:  (Default value = True)
        :param length:  (Default value = -1)

        """

        output_str = ""
        if include_header:
            output_str += "track type=wiggle_0 "
            if include_name:
                output_str += 'name="' + name + '" '
            if include_description:
                output_str += 'description="' + description + '" '
            output_str += "visibility=full"
            output_str += "\n"

        # Note the +1 below. I suspect this is necessary as wig files are
        # 1-based...
        output_str += (
            "fixedStep chrom=chr"
            + str(chr_no)
            + " start="
            + str(displacement + 1)
            + " step=1"
        )

        output_str += "\n"

        # length long array, 0-indexed
        depth_array = self.return_depth(length=length)
        output_str += "\n".join(map(str, depth_array))

        return output_str


if __name__ == "__main__":
    # testing return_depth here:
    test_sites = BindingSites(
        [
            (3, 100, "my name"),
            (102, 1000, "hey this is nahin"),
            (456, 1004, "What's good"),
            (600, 2000, "more random stuff"),
            (4, 20, "What's good"),
        ]
    )
    print(test_sites.return_depth(3000))
    print(len(test_sites.return_depth(3000)))

    # testing overlap mode capabilities
    overlap_sites = BindingSites(
        [
            (3, 100, "my name"),
            (102, 1000, "hey this is nahin"),
            (456, 1004, "What's good"),
            (600, 2000, "more random stuff"),
            (4, 20, "What's good"),
        ],
        overlap_mode=True,
    )
    print(overlap_sites.return_depth(3000))
    print(len(overlap_sites.return_depth(3000)))
    print(test_sites)
    print(overlap_sites)
    merger = "; ".join
    print(
        overlap_sites.overlap_collapse(
            "TopDepthRatio", 1.0, annotation_merger=merger
        )
    )



"""
A file dedicated to useful functions related to position weight matrices that
describe the binding profiles of RBPs, such as scanning an RNA sequence to
detect where the RBP might bind on it. Possibly the file I had the most pleasure
writing, despite its simplicity.

What is a PWM? A PWM is a position weight matrix that summarizes a collection of
strings. In the context of biology, an example could be a collection of
sequences that an RBP was found to bind to protein. Say, for example, that an
RBP called HNRNPK was found to bind to 100 RNA sequences, each of length 5
bases. We could then summarize, for each of those 5 bases, the frequency of
occurrence of a nucleotide base in that position. An example binding profile
could be:

Base / Position
  1   2   3   4   5
A 50  100 100 0   0
G 50  0   0   60  100
C 0   0   0   30  0
T 0   0   0   10  0

Notice how each column adds up to 100, since we had a collection of 100. We can
conclude that this RBP binds to a 5-letter sequence that looks like
(A/G)AA(G/C/T)G, with the distribution shown above (e.g. first letter is equally
likely to be A or G, and so on). Technically, what is shown above is a position
frequency matrix. It is common to normalize the above so each column adds up to
1, to generate a position probability matrix. For some purposes, one might take
the log values of the above too. I think all these things can be called PWMs.
Thus in this project, I have referred to any form of position frequency matrix
(normalized or not) as a PWM. We do not deal with logs here!

Clearly, pwms are useful for making predictions about where an RBP might bind to
an RNA sequence. I did not do extensive research on statistics here to make such
predictions; instead I used a simplest model described on a Wikipedia page,
where one assumes an equiprobable base distribution for prior belief and
independence between motif base positions in terms of nucleotide preference.

Specifically, we decide if an rbp binds to a specific location on an RNA
sequence by associating a "score" with the rbp binding profile on a specific
location, calculated by multiplying all the position weights that show up on the
RNA molecule. The RBP is considered to bind if the score calculated is more than
80% of the highest possible score attainable (cutoff percentage can be adjusted
in config.py).

We could have chosen a number of ways to represent PWM's. Somewhat arbitrarily,
I went with a pwm variable being represented by a dictionary that maps the
letters of the nucleotide bases ("A", "C", etc.) to a list of length equal to
the length of the motif. The list has numbers that represent frequency (not
assumed to be probabilities) of the base at that position. For example,
pwm["T"][6] is the frequency of the base T on the 7th nucleotide of the motif
represented by "pwm".

"""

# TODO: consider defining a class for PWM's for ease of interface


from functools import partial
from pathlib import Path

from .config import GENOME_VERSION, PWM_SCAN_CUT_OFF_PERCENTAGE, RO_DATA_PATH

bases = ["A", "G", "C", "T"]

seqs = {}


def get_human_seq(rna_info):
    """
    Returns the human genomic sequence of the specified chromosome from the
    start to the end coordinate.

    :param rna_info: a dictionary of information regarding the location on the
        human genome from which the nucleotide sequence is sought. Keys provided
        should include:
            'chr_n' : chromosome number
            'start_coord' : start coordinate on the chromosome
            'end_coord' : end coordinate on the chromosome

    """
    chr_no = rna_info["chr_n"]
    chr_start = rna_info["start_coord"]
    chr_end = rna_info["end_coord"]

    if (chr_no, chr_start, chr_end) in seqs:
        return seqs[(chr_no, chr_start, chr_end)]

    chr_files_dir = Path(RO_DATA_PATH) / f"{GENOME_VERSION}-human-genome"
    chr_file = chr_files_dir / f"chr{chr_no}.fa"

    with open(chr_file) as handle:
        # Need to skip the first line, and then a bunch more to reach the line
        # that has the chr_start numbered base. Note that the .fa file comes in
        # lines of 50 nucleotides with "\n" at the end. Admittedly, a part of
        # the below formula was trial and error until it was empirically correct
        # (maybe the -1 part).

        handle.seek(
            4
            + len(str(chr_no))
            + 1
            + 51 * (chr_start // 50)
            + chr_start % 50
            - 1
        )

        # Once at the required position, keep adding lines of bases as long as
        # too short. Then adjust in case we overshot the required length.
        gene = ""
        while len(gene) < chr_end - chr_start:
            gene += handle.readline().strip()
        gene = gene[: chr_end - chr_start]

    seq = gene.upper()
    seqs[(chr_no, chr_start, chr_end)] = seq
    return seq


def product(list_of_numbers):
    """
    Returns a product of numbers.

    :param list_of_numbers: the input list of numbers whose product is sought.

    """
    answer = 1
    for number in list_of_numbers:
        answer *= number
    return answer


def pwm_scan_naive_brute_force(gene, pwm):
    """
    Scans gene, a string of nucleotide bases, for positions where pwm could bind
    using a naive brute force approach. Slower than pwm_scan most of the times,
    but much faster when the degree of freedom (total number of strings
    represented) of pwm is very high (recommended to use this if degree of
    freedom of pwm is > 2000).

    :param gene: a string of nucleotide bases
    :param pwm: a position weight matrix (for format see (no clue))

    """
    len_gene = len(gene)
    len_pwm = len(pwm["A"])
    highest_scores = [max([pwm[b][i] for b in bases]) for i in range(len_pwm)]
    max_score = product(highest_scores)
    cut_off_percentage = 0.80

    cut_off_threshold = cut_off_percentage * max_score

    binding_sites = []
    for i in range(len_gene - len_pwm + 1):
        score = 1
        for j in range(len_pwm):
            score *= pwm[gene[i + j]][j]
            if score < cut_off_threshold:
                break

        if score >= cut_off_threshold:
            binding_sites += [(i, i + len_pwm)]
    return binding_sites


def pwm_degree_of_freedom(pwm):
    """
    Counts the number of binding strings described by a pwm. Can be useful in
    picking between pwm_scan_naive_brute_force and pwm_scan

    :param pwm: a position weight matrix (for format see (no clue))

    """
    # I think that this function can sometimes overestimate the
    # "degree of freedom", because it's just multiplying by the number of bases
    # possibly described at each position by checking if it's above the cutoff
    # percentage of the maximum score at that position, instead of looking
    # holistically. So this function should be considered an approximation.
    len_pwm = len(pwm["A"])
    freedom = 1
    for i in range(len_pwm):
        max_score = max([pwm[b][i] for b in bases])
        freedom *= len(
            [
                1
                for b in bases
                if pwm[b][i] >= PWM_SCAN_CUT_OFF_PERCENTAGE * max_score
            ]
        )
    return freedom


def findall(needle, haystack):
    """
    Yields all the positions of the pattern needle in the string haystack.
    :param needle: substring to be searched for
    :param haystack: the string to be searched in (for needle)

    """
    # Adopted from: https://stackoverflow.com/a/34445090/8551394
    i = haystack.find(needle)
    while i != -1:
        yield i
        i = haystack.find(needle, i + 1)


def pwm_scan(gene, pwm):
    """
    Scans gene, a string of nucleotide bases, for positions where pwm could
    bind. The approach used here is to generate all possible strings of bases
    that the pwm could represent, and use Python's inbuilt substring searching
    algorithm to get all places where the collection of strings could bind on
    the RNA string. Usually fast, but if the pwm represents too many strings
    (e.g. your pwm has a stretch of 15 bases with equi-possible bases
    (think "N"), this algorithm will likely not terminate ever). Consider using
    pwm_scan_naive_brute_force instead in such occasions.

    :param gene: a string of nucleotide bases
    :param pwm: a position weight matrix (for format see (no clue))

    """
    len_pwm = len(pwm["A"])
    cut_off_percentage = PWM_SCAN_CUT_OFF_PERCENTAGE

    possible_seqs = [("", 1)]

    for i in range(len_pwm):
        new_possible_seqs = []
        max_base_score = max([pwm[b][i] for b in bases])
        for seq, score in possible_seqs:
            for base in bases:
                if score * pwm[base][i] / max_base_score >= cut_off_percentage:
                    new_possible_seqs += [
                        (seq + base, score * pwm[base][i] / max_base_score)
                    ]
        possible_seqs = new_possible_seqs

    possible_seqs = [x for x, y in possible_seqs]

    binding_sites = []
    for substring in possible_seqs:
        binding_sites += [(i, i + len_pwm) for i in findall(substring, gene)]
    return binding_sites


def motif_to_pwm(motif, letter_strength=4):
    """
    Converts a motif (a string like "AGCNNNYWS") to a corresponding pwm. In
    making the pwm, letter_strength (defaulted to 4) represents the number of
    times more likely a base is considered to be when represented by the motif
    string at a particular position. For example, in "AGCNNNYWS", the first
    position is A but we don't represent it as a 100% A frequency, instead we
    give A a probability of 4/7 while G,C,T each get a probability of 1/7
    (when letter_strength = 4).

    :param motif: a string representing a nucleotide motif, using standard
        bioinformatics notation (insert ref. link here)
    :param letter_strength: the number of times more likely a base is considered
        to be when represented by the motif string at a particular position.
        (Default value = 4)

    """

    # In practice, the letter_strength might not matter much if it's 4, since
    # our cutoff threshold overall for binding is 80% (so the moment a letter is
    # only 25% it fails). But maybe that's good!

    # I think the above comment talks about use case in RNPFind

    motif = motif.strip()
    motif = motif.replace("U", "T")

    # TODO: Clean up the code below to avoid repeating!

    pwm = {b: [] for b in bases}

    for nucleotide in motif:
        for base in bases:
            if nucleotide == "N":
                pwm[base].append(0.25)
            elif nucleotide == "Y":
                pwm[base].append(
                    letter_strength / (2 * letter_strength + 2)
                    if base in "TC"
                    else 1 / (2 * letter_strength + 2)
                )
            elif nucleotide == "W":
                pwm[base].append(
                    letter_strength / (2 * letter_strength + 2)
                    if base in "TA"
                    else 1 / (2 * letter_strength + 2)
                )
            elif nucleotide == "S":
                pwm[base].append(
                    letter_strength / (2 * letter_strength + 2)
                    if base in "GC"
                    else 1 / (2 * letter_strength + 2)
                )
            elif nucleotide == "K":
                pwm[base].append(
                    letter_strength / (2 * letter_strength + 2)
                    if base in "GT"
                    else 1 / (2 * letter_strength + 2)
                )
            elif nucleotide == "M":
                pwm[base].append(
                    letter_strength / (2 * letter_strength + 2)
                    if base in "AC"
                    else 1 / (2 * letter_strength + 2)
                )
            elif nucleotide == "R":
                pwm[base].append(
                    letter_strength / (2 * letter_strength + 2)
                    if base in "GA"
                    else 1 / (2 * letter_strength + 2)
                )
            elif nucleotide == "D":
                pwm[base].append(
                    letter_strength / (3 * letter_strength + 1)
                    if base != "C"
                    else 1 / (3 * letter_strength + 1)
                )
            elif nucleotide == "H":
                pwm[base].append(
                    letter_strength / (3 * letter_strength + 1)
                    if base != "G"
                    else 1 / (3 * letter_strength + 1)
                )
            elif nucleotide == "B":
                pwm[base].append(
                    letter_strength / (3 * letter_strength + 1)
                    if base != "A"
                    else 1 / (3 * letter_strength + 1)
                )
            else:
                assert nucleotide in bases
                pwm[base].append(
                    letter_strength / (letter_strength + 3)
                    if nucleotide == base
                    else 1 / (letter_strength + 3)
                )

    return pwm


def pwm_summary(pwm):
    """
    Returns a string representing the most likely string represented by the pwm

    :param pwm: a position weight matrix (for format see (no clue))

    """
    len_pwm = len(pwm["A"])

    def get_pwm_base_index(base, index):
        return pwm[base][index]

    return_new = "".join(
        [
            max(bases, key=partial(get_pwm_base_index, index=i))
            for i in range(len_pwm)
        ]
    )

    return return_new


def str_to_pwm(raw_motif_str, is_transpose=False):
    """
    Converts a string containing numbers representing motif frequencies into a
    pwm. is_transpose allows for working with strings oriented differently. Go
    with default if your string has many rows and 4 columns; otherwise set
    is_transpose=True

    :param raw_motif_str: a string containing numbers representing motif
        frequencies
    :param is_transpose: set False if string has many rows and 4 columns,
        otherwise set to True. (Default value = False)

    """
    motif = raw_motif_str.split()
    motif = [float(k) for k in motif]
    base_index = {"A": 0, "G": 1, "C": 2, "T": 3}
    pwm = {b: [] for b in bases}

    for base in bases:
        for i in range(len(motif) // 4):
            if is_transpose:
                pwm[base].append(
                    motif[i + (len(motif) // 4) * base_index[base]]
                )
            else:
                pwm[base].append(motif[i * 4 + base_index[base]])

    return pwm


if __name__ == "__main__":

    # Example usage
    # from user_input import get_user_rna_preference

    # [RNA, RNA_chr_no, RNA_start_chr_coord, RNA_end_chr_coord] = user_input()

    # start = timer()
    # rna_info = RNA, RNA_chr_no, RNA_start_chr_coord, RNA_end_chr_coord

    # RNA_sequence = get_human_seq(
    #     RNA_chr_no, RNA_start_chr_coord, RNA_end_chr_coord
    # )

    # print(RNA_sequence)

    # example_motif = (
    #  '''0.971153846154	0.00961538461538	0.00961538461538	0.00961538461538
    #     0.971153846154	0.00961538461538	0.00961538461538	0.00961538461538
    #     0.971153846154	0.00961538461538	0.00961538461538	0.00961538461538
    #     0.971153846154	0.00961538461538	0.00961538461538	0.00961538461538
    #     0.971153846154	0.00961538461538	0.00961538461538	0.00961538461538
    #     0.971153846154	0.00961538461538	0.00961538461538	0.00961538461538
    #     0.971153846154	0.00961538461538	0.00961538461538	0.00961538461538'''
    # )

    # test_pwm = pwm_str_to_dict(example_motif)

    # sites = pwm_scan(RNA_sequence, test_pwm)
    # for s, t in sites:
    #     print("binding site found!")
    #     print("Position: ", s, "to", t - 1)
    #     print("Sequence: ", RNA_sequence[s:t])

    # print("Number of sites:", len(sites))

    # end = timer()
    # print(end - start, "seconds elapsed")

    # print(pwm_motif_to_dict("(U)(2)"))

    print(
        pwm_scan(
            "CCCCCCTTTTCGCGCGCGCTTTTCGCGCGCGCGCGTTTTTTT", motif_to_pwm("TTTT")
        )
    )
    print(
        pwm_scan_naive_brute_force(
            "CCCCCCTTTTCGCGCGCGCTTTTCGCGCGCGCGCGTTTTTTT", motif_to_pwm("TTTT")
        )
    )
    print(pwm_degree_of_freedom(motif_to_pwm("NNNNN")))
    print(pwm_degree_of_freedom(motif_to_pwm("ACGTGTGTCG")))
    print(pwm_degree_of_freedom(motif_to_pwm("ACGTGTGDDDH")))



"""
This module is dedicated to functions that help generate UCSC Genome Broswer
trackhubs and populate them with binding sites and their annotations, displayed
as simple intervals of sites or as density (dense barchart/histogram) plots.

"""

import glob
import os
from pathlib import Path
from sys import platform

import trackhub

from .binding_analysis_binding_sites import OVERLAP_CONFLICT
from .config import (
    AUTOSQL_PATH,
    GENOME_VERSION,
    UCSC_TRACK_VISIBILITY,
    UCSCTOOL_PATH,
)
from .data_load_functions import (
    column_data,
    data_load_sources_supported_long,
    data_load_sources_supported_short,
)


def populate_local_track_hub(
    overarching_path, rna_info, local_stage, rbp_no_dict, rbp_peaks, rbp=""
):
    """
    Populates a local directory with files conforming to structure required by
    the UCSC specification for displaying trackhubs on their servers.

    It is assumed that all the .bb files have already been generated and are
    stored under a directory in the following structure:
        /<source 1>/<rbp1>.bb
        /<source 1>/<rbp2>.bb
        ...
        /<source 1>/<rbpn1>.bb
        /<source 2>/<rbp1>.bb
        ...
        ...
        /<source m>/<rbpnm>.bb


    This function simply copies the files into a trackhub structure and saves it
    in a specified directory.

    :param overarching_path: Directory under which the .bb files are stored in
        the structure specified above.
    :param rna_info: a dictionary used to represent RNA of interest. It should
        include the 'official_name' and 'chr_n' (chromosome number).
    :param local_stage: the directory to which the local trackhub structured
        files should be saved.
    :param rbp_no_dict: A dictionary containing data sources as keys and the
        number of RBPs discovered (to bind to RNA of interest) by each source as
        value.
    :param rbp_peaks: A dictionary containing data sources as keys and the
        highest number of RBPs discovered to bind to the RNA at one (nucleotide)
        point (just from that data source) as value.
    :param rbp: Name of one RBP of interest, among those binding to the RNA of
        interest (useful only for taking its perspective in competitive /
        cooperative relationship investigation.)

    """

    rna = rna_info["official_name"]
    rna_chr_no = rna_info["chr_n"]

    # # threshold_config_file = open(overarching_path + "threshold_config.txt")
    # with open(
    #     Path(overarching_path) / "threshold_config.txt"
    # ) as threshold_config_file:
    #     _str = threshold_config_file.read()

    assert OVERLAP_CONFLICT == "union"
    hub_name = f"rbps-on-{rna}-{OVERLAP_CONFLICT.lower()}"
    hub, _, _, trackdb = trackhub.default_hub(
        hub_name=hub_name,
        short_label=f"RBPs on {rna}",
        long_label=(
            "RNA binding proteins on the transcript at {rna}. Overlapping"
            " sites were merged into one."
        ),
        genome=GENOME_VERSION,
        email="mnk1@alumni.cmu.edu",
    )

    for filename in glob.iglob(f"{overarching_path}/**/*.bb", recursive=True):
        category = Path(filename).parent.name
        name = Path(filename).name
        # _, _, _, _, _, _, category, name = filename.split("/")
        rbp = name.split("-")[0]
        rbp = rbp.replace(",", "_")
        rbp = rbp.replace("*", "_mut_")
        rbp = rbp.replace("(", "")
        rbp = rbp.replace(")", "")

        visibility = UCSC_TRACK_VISIBILITY
        # TODO: consider options for this for the user (add to Config at least)

        track = trackhub.Track(
            name=rbp + "-" + category + "-binding-sites",
            short_label=rbp + "-" + category,
            long_label=(
                "Binding sites of "
                + rbp
                + " derived from "
                + data_load_sources_supported_long[
                    data_load_sources_supported_short.index(category)
                ]
            ),
            source=filename,
            tracktype="bigBed 9 +",
            itemRgb="on",
            spectrum="on",
            visibility=visibility,
            chromosomes="chr" + str(rna_chr_no),
            # labelFields="", l
            # defaultLabelFields="", l
            # mouseOverField="" l
            # labelFields=",".join(
            #     [column_data[category]["names"][i]
            #     for i in column_data[category]["interest"]]
            # ),
            # defaultLabelFields=",".join(
            #     [column_data[category]["names"][i]
            #     for i in column_data[category]["default_label"]]
            # ),
            # mouseOverField=(
            #     column_data[category]["names"][column_data[category]
            #     ["default_mouse_over"]]
            # ),
            # maxItems=25
        )

        # TODO: Add options for mouse hover views of information

        trackdb.add_tracks(track)

    for filename in glob.iglob(f"{overarching_path}/**/*.bw", recursive=True):
        data_load_source = Path(filename).parent.name
        name = Path(filename).name
        # _, _, _, _, _, _, data_load_source, name = filename.split("/")
        rbp_no = rbp_no_dict[data_load_source]
        rbp_peak = max(rbp_peaks.values())
        rbp_peak = (rbp_peak // 10 + 1) * 10
        visibility = "full"

        # TODO: consider options for this for the user (add to Config at least)

        track = trackhub.Track(
            name=rna + "-" + data_load_source + "-density-plot",
            short_label="00 " + rna + "-density",
            long_label=(
                "Density plot of "
                + str(rbp_no)
                + " RBPs on "
                + rna
                + " using data from "
                + data_load_source.upper()
            ),
            source=filename,
            tracktype="bigWig",
            color="128,0,0",  # TODO: what color ought bigWig density plots be?
            visibility=visibility,
            chromosomes="chr" + str(rna_chr_no),
            viewLimits="0:" + str(rbp_peak),
            maxHeightPixels="128:50:8",
            autoScale="on",
        )

        trackdb.add_tracks(track)

    trackhub.upload.stage_hub(hub, staging=local_stage)

    return hub_name


def prepare_auto_sql(data_load_source):
    """
    Helper function for generating .as files for UCSC visualization. These files
    specify (for each data source) the column title for data that annotates the
    binding sites generated on UCSC genome browser.

    :param data_load_source: The data source for which the .as file should be
        created.

    """
    source_columns_of_interest = range(
        len(column_data[data_load_source]["names"])
    )

    no_of_extra_fields = len(source_columns_of_interest)
    name_of_file = (
        data_load_source
        + "".join([str(c) for c in source_columns_of_interest])
        + ".as"
    )

    file_path = f"{AUTOSQL_PATH}/{name_of_file}"
    template_file_path = f"{AUTOSQL_PATH}/general_template.as"
    try:
        open(file_path, "r").close()
    except FileNotFoundError:
        with open(file_path, "w") as handle:
            # TODO: make an auto generator of auto_sql template files here
            with open(template_file_path, "r") as template_handle:
                template_string = template_handle.read()
                template_string = template_string.replace(
                    "insert_source_name_here", data_load_source
                )

            handle.write(template_string)
            column_names = [
                column_data[data_load_source]["names"][i]
                for i in source_columns_of_interest
            ]
            descriptions = [
                column_data[data_load_source]["descriptions"][i]
                for i in source_columns_of_interest
            ]
            additional_str = ""
            for column, description in zip(column_names, descriptions):
                additional_str += (
                    "\t".join(
                        ["lstring", column + ";", '"' + description + '"']
                    )
                    + "\n"
                )

            additional_str += ")"
            handle.write(additional_str)

    return no_of_extra_fields, name_of_file


def convert_bed_to_bb(overarching_path, data_load_sources):
    """
    Given a directory containing .bed files, converts them all to .bb files and
    saves them. Uses the UCSC bedToBigBed tool as a subroutine.

    :param overarching_path: Directory containing the .bed files in a layout
        similar to as described in the docstring of populate_local_track_hub.

    :param data_load_sources: List of data sources for which the convertion is
        desired.

    """
    debug = False
    if GENOME_VERSION != "hg38":
        raise ValueError("Update this function for this genome version!")

    os_folder = ""
    if platform in ("linux", "linux2"):
        # linux
        os_folder = "linux"
    elif platform == "darwin":
        # OS X
        os_folder = "mac-os"
    assert os_folder

    # starting_working_directory = os.getcwd()
    for data_load_source in data_load_sources:
        no_of_extra_fields, as_file_name = prepare_auto_sql(data_load_source)
        # os.chdir(overarching_path + data_load_source + "/")

        os.system(
            f"for file in {Path(overarching_path) / data_load_source}/*.bed;"
            f" do echo $file; {UCSCTOOL_PATH}/{os_folder}/bedToBigBed"
            f" -as={AUTOSQL_PATH}/{as_file_name}"
            + " type=bed9+"
            + str(no_of_extra_fields)
            + ' "$file" '
            + f'{UCSCTOOL_PATH}/hg38.chrom.sizes "${{file%.*}}".bb; done'
            + (" >/dev/null 2>&1" if not debug else "")
        )

        # os.chdir(starting_working_directory)

    # end of function


def upload_online(local_dir, github_dir):
    """
    Given a local directory containing trackhub-structured files, uploads them
    to the internet so that UCSC Genome Browser can access it.

    This function uploads the local directory to an s3 bucket and assumes that
    s3 credentials have been configured.

    :param local_dir: Local directory with trackhub-structured files.
    :param github_dir: Directory on s3 to save the files on.

    """
    debug = False
    terminator = " >/dev/null 2>&1" if not debug else ""

    os.system("aws s3 sync " + local_dir + " " + github_dir + terminator)


def density_plot(big_storage, rna_info, data_load_sources, overarching_path):

    """
    Generates .wig files containing density information on RBPs binding to RNA
    of interest

    :param big_storage: a dictionary containing data sources as keys and
        Storage instances with the binding sites extracted from
        the corresponding data sources as values.
    :param rna_info: a dictionary specifying the RNA of interest, by specifying
        the name of the RNA and its genomic location on hg38. Keys needed are:
            'official_name' - name of RNA
            'chr_n' - chromosome in hg38 on which it lies
            'start_coord' - the coordinate on which it starts on the chromosome.
    :param data_load_sources: a list of sources for which tthe .wig files
        should be generated, such as 'rbpdb', 'postar', etc.
    :param overarching_path: the directory in which the .wig files should be
        saved.
    :returns: a dictionary containing the number of RBPs that were discovered
        by each data source on the RNA of interest.

    """

    rna = rna_info["official_name"]
    rna_chr_no = rna_info["chr_n"]
    rna_start_chr_coord = rna_info["start_coord"]
    rbp_no_dict = {}
    for data_load_source in data_load_sources:
        storage = big_storage[data_load_source]

        # TODO: check if displacement needs to be shifted by one for all data
        # sources or just RBPDB
        wig_string = storage.print_wig(
            chr_no=rna_chr_no,
            displacement=rna_start_chr_coord - 1,
            include_name=True,
            include_description=True,
            name=rna,
            description=(
                "Density plot of " + str(len(storage)) + " RBPs on " + rna
            ),
            include_header=True,
        )

        rbp_no_dict[data_load_source] = len(storage)

        folder_path = Path(overarching_path) / data_load_source
        filepath = (
            f"{rna}-{data_load_source}-{GENOME_VERSION}-density-plot.wig"
        )

        filepath = folder_path / filepath

        folder_path.mkdir(parents=True, exist_ok=True)

        with open(filepath, "w") as density_plot_wig_file:
            density_plot_wig_file.write(wig_string)

    return rbp_no_dict


def convert_wig_to_bw(overarching_path, data_load_sources):
    """
    Given a directory containing .wig files, converts them all to .bw files and
    saves them. Uses the UCSC wigToBigWig tool as a subroutine.

    :param overarching_path: Directory containing the .wig files in a layout
        similar to as described in the docstring of populate_local_track_hub.

    :param data_load_sources: List of data sources for which the convertion is
        desired.

    """
    if GENOME_VERSION != "hg38":
        raise ValueError("Update this function for this genome version!")

    os_folder = ""
    if platform in ("linux", "linux2"):
        # linux
        os_folder = "linux"
    elif platform == "darwin":
        # OS X
        os_folder = "mac-os"
    assert os_folder

    # starting_working_directory = os.getcwd()
    for data_load_source in data_load_sources:
        # os.chdir(overarching_path + data_load_source + "/")

        os.system(
            f"for file in {overarching_path+data_load_source}/*.wig;"
            f" do echo $file; {UCSCTOOL_PATH}/{os_folder}/wigToBigWig"
            f' "$file" {UCSCTOOL_PATH}/hg38.chrom.sizes '
            '"$file.bw"; done >/dev/null 2>&1'
        )

        # os.chdir(starting_working_directory)



"""
This module contains functions for loading data from the postar database.

"""

import bisect
import os

from .config import ANNOTATION_COLUMN_DELIMITER, POSTAR_PATH

postar_all_column_names = [
    "chrom",
    "chromStart",
    "chromEnd",
    "postarID",
    "nil",
    "strand",
    "rbpName",
    "dataSource",
    "cellType",
    "expSource",
    "postarScore",
]

postar_all_column_descriptions = [
    "chromosome number",
    "start coordinate",
    "end coordinate",
    "POSTAR database ID",
    "not sure",
    "strand",
    "RBP Name",
    "Data Source",
    "Cell type",
    "experimental source",
    "score",
]

postar_columns_of_interest = [3, 7, 8, 9, 10]
postar_default_label_index = [8]
POSTAR_DEFAULT_MOUSE_OVER_INDEX = 9

postar_column_names = [
    postar_all_column_names[i] for i in postar_columns_of_interest
]

postar_column_descriptions = [
    postar_all_column_descriptions[i] for i in postar_columns_of_interest
]


class Query:
    """
    Represents a region of interest within a genome.
    Supports comparison with a line from a file containing details on binding
    sites, in a format much like POSTAR's public database files. Since the file
    is sorted, this allows quick binary search over the file using this class's
    comparison function as a comparator against the lines in the file (to find
    the region of interst, stored in an instance of this class at __init__).

    """

    def __init__(self, query):
        self.query = query

    def __lt__(self, line):
        rna_chr_no, rna_start_chr_coord, rna_end_chr_coord = self.query
        line_parts = line.split()
        # print(line_parts)

        return (line_parts[0], int(line_parts[1]), int(line_parts[2])) > (
            "chr" + str(rna_chr_no),
            rna_start_chr_coord,
            rna_end_chr_coord,
        )

    def __str__(self):
        return str(self.query)


class FileSearcher:
    """
    Class for representing a file as an indexable element. This allows binary
    search over the file using in-built libraries (bisect.bisect)
    """

    def __init__(self, file_pointer):
        self.file_pointer = file_pointer
        self.file_pointer.seek(0, os.SEEK_END)
        self.num_bytes = self.file_pointer.tell() - 1

    def __len__(self):
        return self.num_bytes

    def __getitem__(self, i):
        # TODO: Fix the three corner case bugs inherent in a binary search
        # algorithm like this
        # See: (http://pts.github.io/pts-line-bisect/line_bisect_evolution.html)
        self.file_pointer.seek(i)
        self.file_pointer.readline()

        return self.file_pointer.readline()


def binary_search_populate(file_path, rna_info, debug=False):
    """
    Searches a file containing sorted binding sites for a region of interest.
    Returns the subset of binding sites required as a generator / iterator
    object.

    :param file_path: a file path containing sorted binding sites on the genome.
    :param rna_info: a dictionary containing chromosome number, start, and end
        coordinate to slice out of the genome-wide binding sites file.
    :param debug: prints useful information if set to True, for debugging.
        (Default value = False)
    """
    # TODO: Fix a bug here that causes genes without any data to start
    # collecting the whole genome!!

    rna_chr_no = rna_info["chr_n"]
    rna_start_chr_coord = rna_info["start_coord"]
    rna_end_chr_coord = rna_info["end_coord"]
    query = Query((rna_chr_no, rna_start_chr_coord, rna_end_chr_coord))

    postar_data_file = open(file_path)

    search_file = FileSearcher(postar_data_file)
    to_seek = bisect.bisect(search_file, query)
    postar_data_file.seek(to_seek)
    postar_line_parts = postar_data_file.readline().split()
    # print(postar_line_parts)
    is_found = False
    seen = []
    not_found_counter = 0
    while postar_line_parts:
        if (
            postar_line_parts[0] == "chr" + str(rna_chr_no)
            and int(postar_line_parts[1]) > rna_start_chr_coord
            and int(postar_line_parts[2]) < rna_end_chr_coord
        ):
            is_found = True

            if debug:
                if (postar_line_parts[7]) not in seen:
                    print(";".join(postar_line_parts))
                    seen += [postar_line_parts[7]]

            rbp = postar_line_parts[6]
            start, end = postar_line_parts[1], postar_line_parts[2]
            start = int(start) - rna_start_chr_coord
            end = int(end) - rna_start_chr_coord

            # TODO: Consider reformatting the annotation for visual appeal
            annotation = ANNOTATION_COLUMN_DELIMITER.join(
                [postar_line_parts[i] for i in postar_columns_of_interest]
            )
            yield rbp, start, end, annotation

        elif is_found:
            break
        if not is_found:
            not_found_counter += 1
            if not_found_counter >= 4:
                break
        postar_line_parts = postar_data_file.readline().split()


def postar_data_load(rna_info):
    """
    Returns a generator containing binding sites on input RNA molecule, as
    found on the POSTAR database.
    :param rna_info: dictionary containing input RNA information, such as
        chromosome number, start coordinate, and end coordinate.

    """
    file_path = f"{POSTAR_PATH}/postar-human-RBP-binding-sites-sorted.txt"
    return binary_search_populate(file_path, rna_info)


if __name__ == "__main__":
    test_rna_info = ["MALAT1", 11, 65497688, 65506516]
    postar_data_load(test_rna_info)



"""
This file is responsible for loading RBP-RNA interaction data from the needed
data sources.

Explanation of data structure used in RNPFind  (and thus generated by the
functions of this module):

Top level: big_storage (dict)
Branches (keys): data_load_sources:
'postar', 'attract', 'rbpdb', etc.

Next level:
big_storage[source]
Value: Storage variable (RBP -> RNA intervals mapping)
Branches (keys): RBPs
'HNRNPD', 'HNRNPC', etc.

Next level:
big_storage[source][RBP]
Value: a BindingSites variable,
a bunch of RBP binding sites (intervals)

E.g. To get the binding sites of HNRNPC on
the template RNA as determined experimentally by postar:

big_storage['postar']['HNRNPC']

"""

import sys

from .bind_analysis import BindingSites, Storage
from .config import (
    ANNOTATION_COLUMN_DELIMITER,
    ANNOTATION_ROW_DELIMITER,
    EXPERIMENTAL_BINDING_SITE_ACCEPTABLE_COVERAGE_RATIO,
)
from .data_load_functions import data_load_sources_functions
from .merge_annotation_funcs import generate_merge_func


def load_data(data_load_sources, rna_info: dict):
    """
    Goes over a list of data sources of interest for a particular RNA and
    populates a Storage instance (for each of the data sources) with binding
    sites obtained from the data source. Returns a big_storage variable that
    maps data sources to storage variables that store binding data retrieved
    from the data source.

    :param data_load_sources: a list containing data sources, such as 'rbpdb',
        'postar', etc.
    :param rna_info: dict: a dictionary consisting of information about the RNA
        of interest, such as its name and genomic location.
    :returns: a dictionary mapping data load source to a Storage instance
        containing binding sites obtained from that data source.

    """

    big_storage = {}
    for i, data_load_source in enumerate(data_load_sources):
        print(
            f"Loading binding sites from {data_load_source}", file=sys.stderr
        )

        # TODO: is the merge func still relevant?
        merge_func = generate_merge_func(data_load_source)

        storage_space = Storage(annotation_merge_func=merge_func)
        big_storage[data_load_source] = storage_space
        collected_data = data_load_sources_functions[data_load_source](
            rna_info
        )

        for rbp, start, end, annotation in collected_data:
            if rbp not in storage_space:
                storage_space[rbp] = BindingSites(overlap_mode=True)
            storage_space[rbp].add((start, end, annotation))

        # Now we merge all the binding sites that overlap.

        # TODO: fix the implementation of overlap_collapse so annotations are
        # not lost

        # Get max coverage
        max_coverage = max(
            [
                binding_site.base_cover()
                for rbp, binding_site in storage_space.items()
            ]
        )

        # Filter the allowed amount
        allowed_coverage = (
            EXPERIMENTAL_BINDING_SITE_ACCEPTABLE_COVERAGE_RATIO * max_coverage
        )

        for binding_site in storage_space.values():
            binding_site.overlap_collapse(
                "baseCoverNumber",
                allowed_coverage,
                in_place=True,
                annotation_merger=ANNOTATION_ROW_DELIMITER.join,
            )

    return big_storage


def annotation_to_columns(annotation):
    """
    Defines an annotation-to-columns conversion function for UCSC visualization
    data analysis function. This function takes an annotation as input and
    returns the annotation split into arrays, so that UCSC can display details
    of a binding site as a table.

    :param annotation: input annotation from a binding site.

    """

    rows = annotation.split(ANNOTATION_ROW_DELIMITER)
    array = [tuple(r.split(ANNOTATION_COLUMN_DELIMITER)) for r in rows]
    array = list(set(array))
    len_row = len(array[0])
    no_of_rows = len(array)

    array_of_strings = [
        "______".join([array[i][j] for i in range(no_of_rows)])
        for j in range(len_row)
    ]

    return array_of_strings


keys = ["rbpdb", "attract", "rbpmap", "postar", "custom"]
data_source_annotation_to_columns = {k: annotation_to_columns for k in keys}

if __name__ == "__main__":
    pass



"""
Stores most of the configurable variables for RNPFind for ease of access.
"""

from pathlib import Path

# Building dictionary takes time. Set to false to use an identity function
# instead.
# dictionaryBuild = True
# ncbi_gene_dir = "./website/data/NCBI Gene Info/"

ncbi_gene_files = [
    "Human Genes and Synonyms.xlsx",
    "Saccharomyces_cerevisiae.xlsx",
    "Drosophila_melanogaster.xlsx",
]

# refreshSynonymDict = False

# Loading the data for RBP binding sites on lncRNA, as well as BIOGRID
# protein-protein interactions.
# dataLoad = True


# Analysis takes time. Set to true to analyse:
# analysis_overall_correlation = False
# What is the base pair stringency you want to analyse with?
# analysis_threshold_bps = [10, 15, 30, 50]

# The newer form of analysis: per binding site
# analysis_per_binding_site = False
# Window of bp range to look outside each AUF1 site
# analysis_per_binding_site_window = 56
# Threshold for how much is considered to be competitive
# analysis_per_binding_site_competititive_range = 15

# Some of the variables that take long to load will not be loaded again if this
# parameter is set to False. Namely, neat1_storage, malat1_storage, and
# synonym_dict.

# refreshStorages = True

# Add modified custom data?
# custom_data_add = True

EXPERIMENTAL_BINDING_SITE_ACCEPTABLE_COVERAGE_RATIO = (
    1  # Professor Ihab has decided
)

# Genome version being used
GENOME_VERSION = "hg38"

UCSC_TRACK_VISIBILITY = "dense"

# Used for csv output format
DEFAULT_BASE_STRINGENCY = 30

# This should be False normally. Only set to True if you want to make a
# dedicated directory for your RNA of interest.
# Note that this flag only affects UCSC browser visualization method.
DEDICATED_ANALYSIS = False

# Exactly how strong is a letter when written as a motif, compared with other
# letters? How many times more likely?

# I believe any value from 2-10 makes no difference in practice
RBPDB_MOTIF_PWM_LETTER_STRENGTH = 4
RBPDB_MOTIF_N_REPEAT_REQ = 6

PWM_SCAN_CUT_OFF_PERCENTAGE = 0.80

# This is data loading specific. If you are implementing a data load function,
# make sure your data does not have
# the following substrings in their annotations for binding sites. If they do,
# you should change the strings below!
ANNOTATION_COLUMN_DELIMITER = ",,,,,"
ANNOTATION_ROW_DELIMITER = ";;;;;"

RO_DATA_TAR_NAME = Path(__file__).parent / "all.tar.gz"
RO_DATA_URL = "https://rnpfind.com/ro-data/all.tar.gz"

# Path for all cached data
CACHE_PATH = Path(__file__).parent / "cache"
# Path for pickled data
PICKLE_PATH = f"{CACHE_PATH}/pickles"
# Path for all (input) read-only data
RO_DATA_PATH = Path(__file__).parent / "ro-data"
# Path for RBPDB data
RBPDB_PATH = f"{RO_DATA_PATH}/rbpdb"
# Path for ATTRACT data
ATTRACT_PATH = f"{RO_DATA_PATH}/attract"
# Path for POSTAR data
POSTAR_PATH = f"{RO_DATA_PATH}/postar"
# Path for AutoSQL files
AUTOSQL_PATH = f"{RO_DATA_PATH}/autosql"
# Path for UCSC tools
UCSCTOOL_PATH = "./ucsc-tools"
